import copy

STATUS_OK = 0
STATUS_WARN = 1
STATUS_CRIT = 2
STATUS_UNKNOWN = 3


class Plugin:
    """Base class to implement check plugins.

    Most methods use a check_config dict that mostly defines thresholds.
    It looks like:
    check_config = {
        'warn': '80%',
        'crit': '90%',
        'other_data': 'something'
    }
    """

    def __init__(self, config):
        if config is None:
            config = {}
        self.config = config

    @classmethod
    def get_thresholds(cls, check_config, modifier=None):
        critical_threshold = check_config['crit']
        if modifier:
            critical_threshold = modifier(critical_threshold)
        warning_threshold = check_config['warn']
        if modifier:
            warning_threshold = modifier(warning_threshold)
        return critical_threshold, warning_threshold

    @classmethod
    def _value_to_status_less(cls, value, check_config, modifier=None):
        """Return an error when the value should be less than threshold."""
        critical, warning = cls.get_thresholds(check_config, modifier)
        if value >= critical:
            return STATUS_CRIT
        elif value >= warning:
            return STATUS_WARN
        else:
            return STATUS_OK

    @classmethod
    def _value_to_status_more(cls, value, check_config, modifier=None):
        """Return an error when the value should be more than threshold."""
        critical, warning = cls.get_thresholds(check_config, modifier)
        if value <= critical:
            return STATUS_CRIT
        elif value <= warning:
            return STATUS_WARN
        else:
            return STATUS_OK

    @classmethod
    def _strip_percent_sign(cls, value):
        try:
            return int(value)
        except ValueError:
            return int(value.split('%')[0])

    @classmethod
    def _strip_percent_sign_from_check_config(cls, check_config):
        check_config = copy.deepcopy(check_config)
        check_config['warn'] = cls._strip_percent_sign(check_config['warn'])
        check_config['crit'] = cls._strip_percent_sign(check_config['crit'])
        return check_config


class Check:
    def __init__(self, name, check_func, config):
        self.name = name
        self.check_func = check_func
        self.config = config

    def run_check(self):
        return self.check_func(self.config)


class PsutilPlugin(Plugin):

    def __init__(self, config):
        super().__init__(config)
        try:
            import psutil
            self.psutil = psutil
        except ImportError:
            from .. import DependencyError
            raise DependencyError(self.__class__.__name__, 'psutil',
                                  'psutil', 'python3-psutil')


def get_plugin(plugin_name):
    if plugin_name == 'Load':
        from .load import LoadPlugin
        return LoadPlugin
    if plugin_name == 'Memory':
        from .memory import MemoryPlugin
        return MemoryPlugin
    if plugin_name == 'Disk':
        from .disk import DiskPlugin
        return DiskPlugin
    if plugin_name == 'Command':
        from .command import CommandPlugin
        return CommandPlugin
    if plugin_name == 'Redis':
        from .redis import RedisPlugin
        return RedisPlugin
    if plugin_name == 'Processes':
        from .processes import ProcessesPlugin
        return ProcessesPlugin
    raise ValueError('Unknown plugin {}'.format(plugin_name))


def get_all_plugins():
    from .load import LoadPlugin
    from .memory import MemoryPlugin
    from .disk import DiskPlugin
    from .command import CommandPlugin
    from .redis import RedisPlugin
    from .processes import ProcessesPlugin
    return (LoadPlugin, MemoryPlugin, DiskPlugin, CommandPlugin, RedisPlugin,
            ProcessesPlugin)


def bytes_to_human(n):
    symbols = ('K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y')
    prefix = {}
    for i, s in enumerate(symbols):
        prefix[s] = 1 << (i + 1) * 10
    for s in reversed(symbols):
        if n >= prefix[s]:
            value = float(n) / prefix[s]
            return '%.1f%s' % (value, s)
    return "%sB" % n


def human_to_bytes(size):
    symbols = ('K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y')
    size = size.upper()
    for i, s in enumerate(symbols):
        if s in size:
            base = int(size.split(s)[0])
            return base * (1024 ** (i+1))
    return int(size)



import subprocess
import shlex

from . import Plugin, STATUS_OK, STATUS_WARN, STATUS_CRIT, STATUS_UNKNOWN


class CommandPlugin(Plugin):

    def command(self, check_config):
        p = subprocess.Popen(
            shlex.split(check_config['command']),
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True
        )
        stdout, _ = p.communicate()
        return p.returncode, stdout

    @staticmethod
    def _return_code_to_status(cls, return_code):
        if return_code in (STATUS_OK, STATUS_WARN, STATUS_CRIT):
            return return_code
        return STATUS_UNKNOWN

    @staticmethod
    def config_sample():
        return '''
        # Execute external command
        # Return code is the service status
        Command:
          checks:
            - type: command
              name: check_website
              command: /opt/check_website.sh
        '''



from . import PsutilPlugin


class DiskPlugin(PsutilPlugin):

    def used_percent(self, check_config):
        check_config = self._strip_percent_sign_from_check_config(check_config)
        for part in self.psutil.disk_partitions(all=False):
            part_usage = self.psutil.disk_usage(part.mountpoint).percent
            status = self._value_to_status_less(part_usage, check_config)
            if status > 0:
                return (
                    status,
                    'Partition {} is full at {}%'.format(part.mountpoint,
                                                         part_usage)
                )
        return 0, 'Disk usage correct'

    @staticmethod
    def config_sample():
        return '''
        # Usage of disks
        Disk:
          checks:
            - type: used_percent
              warn: 80%
              crit: 90%
        '''



from . import Consumer


class StdoutConsumer(Consumer):

    def _send(self, service_check):
        print(service_check)

    @staticmethod
    def config_sample():
        return '''
        # Just prints checks on the standard output
        Stdout:
        '''



#!/usr/bin/env python3
"""Health checker

Usage:
  sauna [--level=<lvl>] [--config=FILE]
  sauna sample
  sauna list-checks [--config=FILE]
  sauna (-h | --help)
  sauna --version

Options:
  -h --help      Show this screen.
  --version      Show version.
  --level=<lvl>  Log level [default: warn].
  --config=FILE  Config file [default: sauna.yml].

"""
import logging

from docopt import docopt

import sauna


def main():
    args = docopt(__doc__, version=sauna.__version__)
    conf_file = args['--config']
    logging.basicConfig(
        format='%(asctime)s - %(levelname)-8s - %(message)s',
        datefmt='%Y/%m/%d %H:%M:%S',
        level=getattr(logging, args['--level'].upper(), 'WARN')
    )
    logging.getLogger('requests').setLevel(logging.ERROR)

    if args.get('sample'):
        file_path = sauna.assemble_config_sample('./')
        print('Created file {}'.format(file_path))
    elif args.get('list-checks'):
        for name in sauna.get_checks_name(conf_file):
            print('{}'.format(name))
    else:
        sauna.launch(conf_file)

    logging.shutdown()



"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""
from setuptools import setup, find_packages
from codecs import open
from os import path

from sauna import __version__ as version

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='sauna',
    version=version,
    description='Daemon that runs and reports health checks',
    long_description=long_description,
    url='https://github.com/NicolasLM/sauna',
    author='Nicolas Le Manchet',
    author_email='nicolas@lemanchet.fr',
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: System Administrators',
        'Topic :: System :: Monitoring',
        'Topic :: System :: Systems Administration',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],

    keywords='monitoring health checks nagios shinken',

    packages=find_packages(exclude=['tests']),

    install_requires=[
        'docopt',
        'PyYAML'
    ],

    tests_require=[
        'nose',
        'pep8'
    ],

    entry_points={
        'console_scripts': [
            'sauna = sauna.main:main',
        ],
    },

)
